// 2006-04-30 programming competition
// Tester for the task COIN BOXING
// Author: Ahto Truu
// Tested: GCC 3.2 / MinGW
// Requires WinNT/2k/XP (won't work on Win9x, won't compile on Unix/Linux)

// The solutions of this task are evaluated on a tournament where
// each solution plays several pre-set boards against every other.
// Each pair of solutions plays each board twice, so that each
// of them can be the first mover and also the second mover.
// The list of players is loaded from standard input in the form
//    Name of Player 1
//    command_to_run_player_1
//    Name of Player 2
//    command_to_run_player_2
// Typically, you would run the tournament using a command like
//    pokstester <pokstester.lst >pokstester.log
// The list of boards to play is defined in the constructor of
// the class Tournament, around line 285.

#include <string>
#include <iostream>
#include <algorithm>
#include <vector>
using namespace std;

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>

#include <windows.h>

/**
 * Represents a Win32 process.
 */
struct Process {

    char * m_cmd; // Command line used to execute the process.
    HANDLE m_hpr; // Handle of the process.
    HANDLE m_hwr; // Handle for writing to the stdin of the process.
    HANDLE m_hrd; // Handle for reading from the stdout of the process.

    /** A helper type to pass around the data needed by the reader thread. */
    typedef struct {
        HANDLE hSrc; // Handle to read from.
        char * pBuf; // Buffer to read to.
        DWORD dwMax; // Max number of bytes to read.
        DWORD dwRes; // Number of bytes actually read.
    } READ_DATA;

    /** A thread function to perform the asynchronous read from pipe. */
    static DWORD WINAPI ReadProc(LPVOID p) {
        READ_DATA * prd = (READ_DATA *) p;
        return ReadFile(prd->hSrc, prd->pBuf, prd->dwMax, &prd->dwRes, NULL);
    }

    /** Standard constructor, does not run the process yet. */
    Process(const char * cmd) {
        m_cmd = strdup(cmd);
        m_hpr = 0;
        m_hwr = 0;
        m_hrd = 0;
    }

    /** Standard destructor, kills the process if still running. */
    ~Process() {
        free(m_cmd);
        if (!dead())
            kill();
        if (m_hpr)
            CloseHandle(m_hpr);
    }

    /** Attempts to run the process, returns false on any errors. */
    bool exec() {
        STARTUPINFO si = {sizeof(si)};
        SECURITY_ATTRIBUTES sa = {sizeof(sa), NULL, TRUE};
        if (!CreatePipe(&si.hStdInput, &m_hwr, &sa, 0))
            return false;
        if (!CreatePipe(&m_hrd, &si.hStdOutput, &sa, 0))
            return false;
        if (!DuplicateHandle(GetCurrentProcess(), GetStdHandle(STD_ERROR_HANDLE),
                GetCurrentProcess(), &si.hStdError, 0, TRUE, DUPLICATE_SAME_ACCESS))
            return false;
        si.dwFlags |= STARTF_USESTDHANDLES;

        PROCESS_INFORMATION pi;
        if (!CreateProcess(NULL, m_cmd, NULL, NULL, TRUE, 0, NULL, NULL, &si, &pi))
            return false;
        m_hpr = pi.hProcess;
        CloseHandle(pi.hThread);

        return true;
    }

    /** Checks wheteher the process has been executed. */
    bool born() {
        return m_hpr;
    }

    /** Checks whether the process has finished. */
    bool dead() {
        if (!m_hpr)
            return false;
        return (WAIT_OBJECT_0 == WaitForSingleObject(m_hpr, 0));
    }

    /** Forcefully terminates the process. */
    void kill() {
        if (m_hwr) {
            CloseHandle(m_hwr);
            m_hwr = 0;
        }
        if (m_hrd) {
            CloseHandle(m_hrd);
            m_hrd = 0;
        }
        if (m_hpr)
            TerminateProcess(m_hpr, 0);
    }

    /** Returns the CPU time consumed by the process so far, or -1 on any errors. */
    long time() {
        if (!m_hpr)
            return 0;
        FILETIME ct, et, kt, ut;
        if (!GetProcessTimes(m_hpr, &ct, &et, &kt, &ut))
            return -1;
        int64_t ktt = (int64_t) kt.dwHighDateTime << 32 + kt.dwLowDateTime;
        int64_t utt = (int64_t) ut.dwHighDateTime << 32 + ut.dwLowDateTime;
        return (ktt + utt) / 10000; // to milli-seconds
    }

    /** Sends data to the process, returns false on any errors. */
    bool send(const char * s) {
        if (!m_hwr)
            return false;
        DWORD n = strlen(s), nn = 0;
        if (!WriteFile(m_hwr, s, n, &nn, NULL) || nn < n)
            return false;
        return true;
    }

    /** Receives data from the process, returns false on any errors or on timeout. */
    bool recv(char * s, int n, long t) {
        if (!m_hrd)
            return false;
        READ_DATA rd = {m_hrd, s, n - 1, 0};
        HANDLE ht = CreateThread(NULL, 0, ReadProc, &rd, 0, NULL);
        if (!ht)
            return false;
        if (WAIT_OBJECT_0 != WaitForSingleObject(ht, t))
            TerminateThread(ht, 0);
        DWORD rt = 0;
        GetExitCodeThread(ht, &rt);
        CloseHandle(ht);
        if (rt && rt != STILL_ACTIVE) {
            s[rd.dwRes] = 0;
            return true;
        }
        return false;
    }
};

/**
 * Represents the game board.
 */
struct Board {

    static const int max_r = 4;

    int m_r; // Number of rows on the board.
    int m_l[max_r]; // Length of each row.
    int m_a[max_r]; // Coins of the first player.
    int m_b[max_r]; // Coins of the second player.

    /** Standard constructor, generates a new random board. */
    Board(int n1, int n2, int n3 = 0, int n4 = 0) {
        m_r = 0;
        if (n1 > 0) m_l[m_r++] = n1;
        if (n2 > 0) m_l[m_r++] = n2;
        if (n3 > 0) m_l[m_r++] = n3;
        if (n4 > 0) m_l[m_r++] = n4;
        for (int i = 0; i < m_r; ++i) {
            m_a[i] = 1; m_b[i] = m_l[i];
        }
        assert(m_r <= max_r);
    }

    /** Using default copy constructor. */
    //Board(const Board & b)

    /** Using default standard destructor. */
    //~Board()

    /** Checks whether the player p has lost. */
    bool lost(int p) {
        switch (p) {
            case 0:
                for (int i = 0; i < m_r; ++i)
                    if (m_a[i] > 1 || m_a[i] + 1 < m_b[i])
                        return false;
                return true;
            case 1:
                for (int i = 0; i < m_r; ++i)
                    if (m_b[i] - 1 > m_a[i] || m_b[i] < m_l[i])
                        return false;
                return true;
            default:
                assert(false);
        }
    }

    /** Applies the player p's move "i k", returns false if move invalid. */
    bool move(int p, int i, int k) {
        if (i == 0 && k == 0)
            return lost(p);
        if (i < 1 || i > m_r)
            return false;
        if (k == 0)
            return false;
        switch (p) {
            case 0:
                m_a[i - 1] += k;
                return (m_a[i - 1] >= 1 && m_a[i - 1] < m_b[i - 1]);
            case 1:
                m_b[i - 1] -= k;
                return (m_b[i - 1] > m_a[i - 1] && m_b[i - 1] <= m_l[i - 1]);
            default:
                assert(false);
        }
    }

    /** Dumps the board, for debugging. */
    void dump(ostream & str) {
        for (int i = 0; i < m_r; ++i) {
            str << " " << m_a[i] << "." << m_b[i] << "." << m_l[i];
        }
        str << endl;
    }
};

/**
 * Represents a player in the tournament.
 */
struct Player {

    string m_name; // Name of the player.
    string m_prog; // Command line to run.
    int m_score;   // Current score.

    /** Standard constructor. */
    Player(const string & name, const string & prog) {
        m_name = name;
        m_prog = prog;
        m_score = 0;
    }

    /** Comparison operator for sorting. */
    bool operator< (const Player & p) const {
        if (m_score == p.m_score)
            return m_name < p.m_name;
        else
            return m_score > p.m_score;
    }
};

/**
 * Represents a tournament.
 */
struct Tournament {

    vector<Player> m_list; // Players in the tournament.
    vector<Board> m_games; // The list of boards each pair of players will play.
    int m_limit; // Time limit per game.
    int m_extra; // Additional time for clean-up.

    /** Standard constructor. */
    Tournament(istream & str, int t) {
        string name, prog;
        while (getline(str, name) && getline(str, prog))
            m_list.push_back(Player(name, prog));
        cerr << "Loaded " << m_list.size() << " players." << endl;
        m_games.push_back(Board(42, 43));
        m_games.push_back(Board(67, 68, 69));
        m_games.push_back(Board(92, 93, 94, 95));
        cerr << "Created " << m_games.size() << " boards." << endl;
        m_limit = t; m_extra = 100;
        cerr << "Time limit for all games: " << m_limit << " ms." << endl;
    }

    /** Sends a board to a process acting as player p. */
    bool send(Process & pp, const Board & b, int p) {
        char s[50];
        int n = sprintf(s, "%d ", b.m_r);
        for (int i = 0; i < b.m_r; ++i)
            n += sprintf(s + n, "%d ", b.m_l[i]);
        n += sprintf(s + n, "%d\r\n", p + 1);
        return pp.send(s);
    }

    /** Sends a move to a process. */
    bool send(Process & pp, int i, int k) {
        char s[50];
        sprintf(s, "%d %d\r\n", i, k);
        return pp.send(s);
    }

    /** Receives a move from a process. */
    bool recv(Process & pp, int & i, int & k) {
        char s[50];
        if (!pp.recv(s, sizeof(s), m_limit))
            return false;
        int n;
        if (sscanf(s, "%d %d %n", &i, &k, &n) != 2 || n < strlen(s))
            return false;
        return true;
    }

    /** Plays one move from player p. */
    int move(Board & b, Process * pp[], int p) {
        int i, k;
        if (!pp[p]->born())
            if (!pp[p]->exec() || !send(*pp[p], b, p)) {
                cout << "Player " << p + 1 << ": could not start" << endl;
                return p;
            }
        if (!recv(*pp[p], i, k)) {
            cout << "Player " << p + 1 << ": could not get move" << endl;
            return p;
        }
        if (pp[p]->time() > m_limit + m_extra) {
            cout << "Player " << p + 1 << ": time limit exceeded" << endl;
            return p;
        }
        if (!b.move(p, i, k)) {
            cout << "Player " << p + 1 << ": invalid move: " << i << " " << k << endl;
            return p;
        }
        cout << "Player " << p + 1 << ": valid move: " << i << " " << k << endl;
        if (!pp[1 - p]->born())
            if (!pp[1 - p]->exec() || !send(*pp[1 - p], b, 1 - p)) {
                cout << "Player " << (1 - p) + 1 << ": could not start" << endl;
                return 1 - p;
            }
        if (!send(*pp[1 - p], i, k)) {
            cout << "Player " << (1 - p) + 1 << ": could not send move" << endl;
            return 1 - p;
        }
        if (i == 0 && k == 0) {
            Sleep(m_extra);
            if (!pp[p]->dead())
                return p;
            if (!pp[1 - p]->dead())
                return 1 - p;
            return 2;
        }
        return 3;
    }

    /** Plays one game. */
    void game(const Board & bb, Player & p0, Player & p1) {
        cout << endl << "Game " << p0.m_name << " vs " << p1.m_name << endl;
        Board b(bb);
        cout << "Starting position:"; b.dump(cout);
        Player * p[] = {&p0, &p1};
        Process pp0(p0.m_prog.c_str());
        Process pp1(p1.m_prog.c_str());
        Process * pp[] = {&pp0, &pp1};
        for (int cp = 0; ; cp = 1 - cp) {
            switch (move(b, pp, cp)) {
                case 0:
                    cout << "Game over: " << p0.m_name << " made a fault" << endl;
                    p1.m_score += 3;
                    return;
                case 1:
                    cout << "Game over: " << p1.m_name << " made a fault" << endl;
                    p0.m_score += 3;
                    return;
                case 2:
                    cout << "Game over: " << p[1 - cp]->m_name << " won" << endl;
                    p[1 - cp]->m_score += 3;
                    p[cp]->m_score += 1;
                    return;
                default:
                    // normal move processed, just keep going
                    continue;
            }
        }
    }

    /** Runs the tournament. */
    void play() {
        const int gg = m_list.size() * (m_list.size() - 1) * m_games.size();
        int g = 0;
        for (int i = 0; i < m_list.size(); ++i)
            for (int j = i + 1; j < m_list.size(); ++j)
                for (int b = 0; b < m_games.size(); ++b) {
                    cerr << "Game " << ++g << " of " << gg << "... ";
                    game(m_games[b], m_list[i], m_list[j]);
                    cerr << "done." << endl;
                    cerr << "Game " << ++g << " of " << gg << "... ";
                    game(m_games[b], m_list[j], m_list[i]);
                    cerr << "done." << endl;
                }
        cout << endl << "Final scores" << endl;
        sort(m_list.begin(), m_list.end());
        for (int i = 0; i < m_list.size(); ++i)
            cout << m_list[i].m_name << " " << m_list[i].m_score << endl;
    }
};

int main() {
    Tournament(cin, 1000).play();
    return 0;
}
